package jfftw;

/**      Class to handle Complex Data arrays in either Real or
 *       Fourier space. This is the simplest fft interface
 *       with simple support for one, two and three dimensional
 *       complex data with, almost, all indexing problems covered
 *       by the suppoed methods.
 *       <p>
 *       Novice users should start here, and use the more complex
 *       <code>RealDataArray</code> or the low-level <code>FFTW</code>
 *       if they want need either more efficient interface or have data
 *       formated other than the interleaved complex implemented here.
 *       <p>
 *       @author Will Hossack, The University of Edinburgh
 */
public class ComplexDataArray extends DataArray {

    private static FFTWComplex fftwc = new FFTWComplex(fftw);

    /**    Constructor to form a three-dimensional
     *     ComplexDataArray, which defaults to <code>FFTW.REAL</code> space.
     *     <p>
     *     All dete is pre-initialsied to zero.
     *     @param w the width (primary) dimension
     *     @param h the height (secondard) dimension
     *     @param d the depth (third) dimension
     */
    public ComplexDataArray(int w, int h, int d) {
	setDimensions(w,h,d);
	setSpace(FFTW.REAL);
	dataBuffer = new double[2*w*h*d];
    }

    /**    Constructor to form a two-dimensional
     *     ComplexDataArray, which defalus to <code>FFTW.REAL</code> space.
     *     @param w the width (primary) dimension
     *     @param h the height (secondard) dimension
     */
    public ComplexDataArray(int w, int h) {
	this(w,h,1);
    }

    /**    Constructor to form a one-dimensional
     *     ComplexDataArray, which defaults to <code>FFTW.REAL</code> space.
     *     @param w the width (primary) dimension
     */
    public ComplexDataArray(int w) {
	this(w,1,1);
    }

    /**    Form a ComplexDataArray with all parametrs taken
     *     from the specified ComplexDataArray. The underlying
     *     data will also be copied.
     *     @param data the ComplexDataArray
     */
    public ComplexDataArray(ComplexDataArray data) {
	setDimensions(data.getWidth(),
		      data.getHeight(),data.getDepth());
	setSpace(data.getSpace());
	dataBuffer = data.getDataBuffer().clone();
    }


    /**   Form a general ComplexDataArray with specified dimensional 
     *    and supplied data array, which is NOT copied, it is
     *    used internally.
     *    <p>
     *    Use with care, you must format the supplied data in the 
     *    correct row interleaved format. If you are using this
     *    in a main program, consided using the lowever level
     *    <code>FFTWComplex</code> class.
     * 
     *     @param w the width (primary) dimension
     *     @param h the height (secondard) dimension
     *     @param d the depth (third) dimension
     *     @param da the data array.
     *     @throws <code>IllegalArgumentException</code> if length of
     *                   data array does not match parameters.

     */
    public ComplexDataArray(int w, int h, int d, double[] da ) {
	setDimensions(w,h,d);
	setSpace(FFTW.REAL);
	if (da.length != 2*w*h*d) {     // Have been send junk
	    throw new IllegalArgumentException( 
	"ComplexDataArray: length of data array does not match parameters");

	}
	dataBuffer = da;
    }

    /**    Form a ComplexData array with real part of the data
     *     coming from a <code>RealDataArray</code>, 
     *     the imaginary part being zero.
     *     This is only valid for REAL space data.
     *     @param data the source RealDataArray
     *    @throws <code>IllegalArgumentException</code> if 
     *    suppied ReadDataArray is not in real space.
     */
    public ComplexDataArray(RealDataArray data) {
	this(data.getWidth(),data.getHeight(),data.getDepth());
	if (data.getSpace() == FFTW.REAL) {    // Real space only
	    for(int i = 0; i < data.length(); i++) {
		setComplex(i,data.getDouble(i),0.0);
	    }
	}
	else {
	    throw new IllegalArgumentException( 
        "ComplexDataArray: called with RealDataArray in Fourier space.");
	}
    } 
	
    
    /**    Form a deep clone of current ComplexDataArray, including data.
     *     @return <code>ComplexDataArray</code> cloned of current.
     */
    public ComplexDataArray clone() {
	return new ComplexDataArray(this);
    }
    
    /**    Method to get a Complex element in one-dimensional case. 
     *     This implements the abstarct methods defined 
     *     in <code>DataArray</code> and handles the internal indexing.
     *     <p>
     *     For a multi-dimensional array this treat the databuffer as 
     *     interleaved Complex array of length 
     *     <code>width*height*depth</code> allowing more efficiency 
     *     one-dimensional access. Use the feature with care since it
     *     bypasses the package array bound checking. 
     *     @param i the element index
     */
    public Complex getComplex(int i) {
	i *= 2;
	return new Complex(dataBuffer[i], dataBuffer[i+1]);
    }

    /**    Method to get the double value of the element on one-dimensions
     *     where the Complex real conversion is controlled by 
     *     <code>setConversion()</code>, which defaults to 
     *     <code>Complex.MODULUS</code>.
     *     @param i the element index.
     */
    public double getDouble(int i) {
	if (conversionFlag == Complex.REAL) {    // Trap real/imag
	    return dataBuffer[2*i];
	} 
	else if (conversionFlag == Complex.IMAG) {
	    return dataBuffer[2*i + 1];
	}
	else {                                    // Do it the hard way....
	return getComplex(i).getDouble(conversionFlag);
	}
    }

    /**    Method to set a Complex point in one-dimension with two doubles.
     *     @param i the element index
     *     @param a the real part 
     *     @param b the imaginary part
     */
    public void setComplex(int i, double a, double b) {
	i *= 2;
	dataBuffer[i] = a;
	dataBuffer[i + 1] = b;
    }

    /**   Method to set the Real part of an element with a double,
     *    the imaginary part is not modified.
     *    @param i the element index
     *    @param a the double value
     */
    public void setDouble(int i, double a) {
	dataBuffer[2*i] = a;
    }


    /**    Multiply an element by a double
     *     @param i the element index
     *     @param a the multiplier
     */
    public void mult(int i, double a) {
	i *= 2;
	dataBuffer[i] *= a;
	dataBuffer[i+1] *= a;
    }

    /**    Multiply an element by a Complex spefied
     *     @param i the element index
     *     @param a the real part
     *     @param b the imaginary part
     */
    public void mult(int i, double a, double b) {
	i *= 2;                   // Real element location
	int ii = i + 1;           // Imaginary element location
	
	//                  Do the multiplication in lowlevel for reasonable
	//                  efficiency.
	double x = dataBuffer[i];
	double y = dataBuffer[ii];
	dataBuffer[i] = a*x - b*y;
	dataBuffer[ii] = b*x + a*y;
    }


    /**    Add a double to real part of element
     *     @param i the element index
     *     @param a to be added to real part
     */
    public void add(int i, double a) {
	i *= 2;
	dataBuffer[i] += a;
    }

    /**    Add a Complex to specified
     *     @param i the element index
     *     @param a the real part
     *     @param b the imaginary part
     */
    public void add(int i, double a, double b) {
	i *= 2;
	dataBuffer[i] += a;
	dataBuffer[i+1] += b;
    }


    /**    Form the conjugate of the element on one-domensions
     *     by megating the imaginary parts.
     *     @param i the element to be conjugated. 
     */
    public void conjugate(int i) {
	int ii = 2*i + 1;
	dataBuffer[ii] = -dataBuffer[ii];
    }
	

    /**   Metod to form the Complex conjugate to the
     *    whole <code>ComplexDataArray</code>.
     */
    public void conjugate() {
	//      Negate every second element of the databuffer
	//      which is where the imaginary parts are held.
	for(int i = 1; i < dataBuffer.length; i += 2 ) {
	    dataBuffer[i] = -dataBuffer[i];
	}
    }
    

    /**    Method to get underlying FFTWComplex
     *     @return <code>FFTWComplex</code> used to do transforms.
     */
    public FFTWComplex getFFTW() {
	return fftwc;
    }


    /**    Method to take in-place fft. If the space is REAL, then forward fft
     *     is taken, else inverse. Normalsiation by total data size will occur
     *     on the inverse transform provided that the set normalsiation flag
     *     is set to <code>true</code>
     *     @see #setNormalisation 
     */
    public void fourier() {
	int direction = 0;

	//             Decide on direction and reset space
	if (getSpace() == FFTW.REAL) {
	    direction = 1;
	    setSpace(FFTW.FOURIER);
	}
	else {
	    direction = -1;
	    setSpace(FFTW.REAL);
	}
	
	//            Do in-place fft (of the correct dimension)
	if (depth > 1) {
	    dataBuffer = fftwc.threeDimensional(width,height,depth,
						dataBuffer,direction,true);
	}
	else if (width > 1) {
	    dataBuffer = fftwc.twoDimensional(width,height,
						dataBuffer,direction,true);
	}
	else {
	    dataBuffer = fftwc.oneDimensional(dataBuffer,direction,true);
	}
	    
	/*    If we have gone back to REAL space, and normalisation is
	 *    set, divide through by number of data points.
	 */
	if (getSpace() == FFTW.REAL && getNormalisation() ) normaliseDB();
    }



    /**     Methdo to apply a +/1 checker patterns to the Complex
     *      data, used to form centred FFTW.
     */
    public void checker() {
        	    
	for(int k = 0; k < depth; k++) {
	    int kstart = k%2;
	    int planeStart = k*width*height;      // Start of plane
	    for(int j = 0; j < height; j++) {
		int jstart = (j + kstart)%2;                   // alternate 0/1 start
		int lineStart = j*width;            // Start of the line
		for(int i = jstart; i < width; i += 2) {
		    int realPart = 2*(lineStart + i); // Location on real
		    int imagPart = realPart + 1;      // Location of imaginary
		    dataBuffer[realPart] = -dataBuffer[realPart];       // Invert real
		    dataBuffer[imagPart] = -dataBuffer[imagPart];       // Invert imag
		}
	    }
	}
    }



}
