package jfftw;
import java.io.*;

/**      Basic class for Java access to the <a href="http://www.fftw.org/">fftw3</a>
 *       library to take  Fourier Transforms or one-dimensional, two-dimensional
 *       or three-dimensional data held in <code>double[]</code> arrays.
 *       <p>
 *       This class set package defaults, manages the wisdom information
 *       and loads the sharable library. Actual numerical use of this package
 *       is via the extending classes. 
 *
 *       @author Will Hossack, The University of Edinburgh
 *       @version 1.2
 */
public class FFTW extends Object {

    /**       Flag for Real space 
     */
    public final static int REAL = 0x1;

    /**       Flag for Fourier space 
     */
    public final static int FOURIER = 0x2;

    /**       Flag for forward FFT
     */    
    public final static int FORWARD = 1;

    /**       Flag for backward (on inverse) FFT
     */
    public final static int BACKWARD = -1;

    /**      Flag for exhaustive plan search (very slow)
     */
    public final static int EXHAUSTIVE = (1 << 3);

    /**      Flag for patient plan search (medium slow)
     */
    public final static int PATIENT = (1 << 5);

    /**     Flag for quick and dirty plan search (fast, but no optimal)
     *      <p>
     *      Default for jttfw but NOT the normal FFTW default
     */
    public final static int ESTIMATE = (1 << 6);

    
    /**    PrintStream for warning and error messages. Defaults to 
     *     <code>System.err</code>.
     */
    protected static PrintStream errorStream = System.err;
    
    /**    Integer to set to the current plan in use by the package.
     *     Defaults to FFTW.ESTIMATE.
     *     
     */
    protected int planFlag = FFTW.ESTIMATE;

    /**     Boolean to deterime in the sharable library has been loaded
     *      sucessfully.
     */
    protected static boolean loaded;

    private static String systemWisdomFile = "/etc/fftw/wisdom";

   

    /**      Default constuctor, do nothing but load the sharable
     *       library.
     */
    public FFTW() {
    }

    /**      Conststrcutor to for a new FFTW, being a copy of the
     *       specifed one. The PlanFlag is copied.
     *       @param fft the FFTW 
     *       @since 1.1
     */
    public FFTW(FFTW fft) {
	setPlanFlag(fft.getPlanFlag());          // Copy the plan flag
    }

    /**      Constructor with flag to optionally load the
     *       system wisdom file. If load fails there is a non-fatal warning.
     *       @param systemWisdom if true default system wisdom file
     */
    public FFTW(boolean systemWisdom) {
	if (systemWisdom) {
	    if (!loadWisdom()) {
		errorStream.println("FFTW warning: Failed to load system wisdom");
	    }
	}
    }
	 

    /**      Constructor to load specifed wisdom file.  
     *       If load fails there is a non-fatal warning.
     *       @param wisdomFile the FFTW wisdom file
     */
    public FFTW(File wisdomFile) {
	if (!loadWisdom(wisdomFile)) {
	    errorStream.println("FFTW warning: Failed to load wisdom " + 
			       wisdomFile.getName());
	}
    }

    /**     Constuctor to load specifed wisdom file specified by name.
     *      If load fails there is a non-fatal warning.
     *      @param wisdomFileName the name of the wisdom file
     */
    public FFTW(String wisdomFileName) {
	if (!loadWisdom(wisdomFileName)) {
	    errorStream.println("FFTW warning: Failed to load wisdom " + 
			       wisdomFileName);
	}
    }

    /**     Method to set the plan flag, default is ESTIMATE
     *      @param flag the plan flag
     */
    public void setPlanFlag(int flag) {
	planFlag = flag;
    }

    /**    Method to add to the plan flag
     *     @param addition the addition to add
     */
    public void addPlanFlag(int addition) {
	planFlag += addition;
    }

    /**     Method to get the plan flag
     *      @return int the plan flag
     */
    public int getPlanFlag() {
	return planFlag;
    }


    /**    Method to reset the default errorStream for package. The default
     *     is System.err
     *     @param es the error PrintStream
     *     @since 1.2 
     */
    public void setErrorStream(PrintStream es) {
	errorStream = es;
    }

    /**    Method to reset the system widsom file location. Note this
     *     does not load the file.
     *     @param fileName name of default widsom file
     */
    public void setSystemWisdom(String fileName) {
	systemWisdomFile = new String(fileName);
    }

    /**    Boolean method to load system widsom file
     *     @return boolean true for success, false for failure
     */
    public boolean loadWisdom() {
	return loadWisdom(new File(systemWisdomFile));
    }

    /**   Boolean method to load wisdom file from specifed File
     *    @param file the wisdom file
     *    @return boolean true for sucess, false for failure
     */
    public boolean loadWisdom(File file) {
	return nativeLoadWisdomFromFile(file.getAbsolutePath());
    }


    /**  Boolean method to load (or add) wisdom from a String.
     *   @param wisdom the wisdom String
     *   @return boolean true for success, false for failure
     */
    public boolean loadWisdom(String wisdom) {
	return nativeLoadWisdomFromString(wisdom);
    }

    /**  Private native method load wisdom from file
     */
    private native boolean nativeLoadWisdomFromFile(String fileName);


    /**  Private native method load wisdom from String
     */
    private native boolean nativeLoadWisdomFromString(String wisdom);

    /**   Void method to clear the wisdom information
     */
    public void clearWisdom() {
	nativeClearWisdom();
    }

    /**  Private native to acually clear the wisdom information
     */
    private native void nativeClearWisdom();


    /** Boolean method to export the current wisdom to a file
     *  @param file the output file
     *  @return <code>boolean</code> true if successful, else false
     */
    public boolean exportWisdom(File file) {
	return nativeExportWisdomToFile(file.getAbsolutePath());
    }

    /**  Private native method to export ther wisdom to a file
     */
    private native boolean nativeExportWisdomToFile(String fileName);

    
    /**   Method to get the wisdom information as a String 
     *    @return <code>String</code> returns the FFTW wisdom
     *           entry in a foramatted String.
     */
    public String getWisdom() {
	return nativeGetWisdom();
    }

    /**  Private native method to get the wisdom as a Java string
     */
    private native String nativeGetWisdom();



    /**  Static method to read a Wisdom file into a String
     *   @param file the wisdom file
     *   @return String the wisdom as a String 
     */
    public static String readWisdom(File file) {
	
	StringBuilder buffer = new StringBuilder();
	String line;
	try {
	    BufferedReader in = new BufferedReader(new FileReader(file));
	    while((line = in.readLine()) != null) {
		buffer.append(line);               // Append the line
		buffer.append('\n');               // Append newline
	    }
	    in.close();
	} catch (Exception e) {
	    errorStream.println("FFTW.readWisdom: failed to read : " +
			       file.getName());
	}
	return buffer.toString();
    }

    /**  Static method to write wisdom String to a file
     *   @param wisdom the wisdom String
     *   @param file the output file
     */
    public static boolean writeWisdom(String wisdom, File file) {
	
	try {
	    BufferedWriter out = new BufferedWriter(new FileWriter(file));
	    out.write(wisdom,0,wisdom.length());
	    out.close();
	} catch (Exception e) {
	    errorStream.println("FFTW.writeWisdom: failed to write : " +
			       file.getName());
	    return false;
	}
	return true;
    }


    /**      Force the shared library to be loaded here
     *       Under Linux this libary must be libjfftw3.so and either
     *       located in /usr/java/jdk<version>/jre/lib/i386/
     *       or under locations with env variable
     *       LD_LIBRARY_PATH.
     */
    static {
	try {
	    System.loadLibrary("jFFTW3");
	    loaded = true;
	} catch (UnsatisfiedLinkError e) {
	    throw new UnsatisfiedLinkError("Fail to load jfftw3 library : " +
					   e.getMessage() + 
					   "\nCheck LD_LIBRARY_PATH");
	}
	//   loaded = false;             // Failed to load library
	
    }
}

    
