package jfftw;
import java.io.File;

/**       Class to implement simple one-off Complex to Complex Fourier
 *        transforms using jfftw as a native library.
 *        <p>
 *        A new plan is created for each FFT. This is sub-optimal for a large number of 
 *        FFTs, but reasonable for ``one-off''.
 *        @author Will Hossack, The Univesrity of Edinburgh
 *        @version 1.2
 *
 */
public class FFTWComplex extends FFTW {

    /**      Default constuctor
     */
    public FFTWComplex() {
    }
    
    /**      Conststrcutor to for a FFTWComplex from a FFTW, 
     *       The PlanFlag is also copied.
     *       @param fft the FFTW 
     *       @since 1.1
     */
    public FFTWComplex(FFTW fft) {
	super(fft);
    }
    

    /**      Constuctor to optionally load system wisdom file
     *       @param systemWisdom if true loads system wisdom file
     */
    public FFTWComplex(boolean systemWisdom) {
	super(systemWisdom);
    }

    /**     Constructor to load specifed wisdom file
     *      @param wisdomFile
     */
    public FFTWComplex(File wisdomFile) {
	super(wisdomFile);
    }

    /**      Constructor to load specifed wisdom file by name
     *       @param wisdomFile
     */
    public FFTWComplex(String wisdomFile) {
	super(wisdomFile);
    }

    /**      Method to take one-dimensional Complex FFT with the data supplied
     *       in a one-dimensional interleaved <code>double</code> array with real parts in 
     *       even elements, and imaginary in the odd elements.
     *       @param data the data to be transformed
     *       @param dirn direction +1 for forward, -1 to backward
     *       @param overwrite if true data is overwritten with FFT, else a
     *              new array is returned.
     *       @return <code>double[]</code> the FFTed array (if overwrite = true, this will
     *               overwrite the input array).
     */
    public double[] oneDimensional(double data[], int dirn, boolean overwrite) {
	
	if (data == null || data.length == 0) {  // Check sanity
	    throw new NullPointerException("FFTComplex.oneDimensional: called with null or zero length data array.");
	}

	double out[] = null;                     // The output array
	if (overwrite) {                         // Inplace transform
	    out = data;
	}
	else {                                   // Make new array of same length
	    out = new double[data.length];
	}

	if (loaded) {
	    nativeOneDimensional(data,out,dirn,getPlanFlag());    // Call native method to do work
	}
	else {
	    errorStream.println("FFTWComplex.oneDimensional: library not loaded.");
	}
	return out;                                           // Return output   
    }


    /**       nativeOneDimensional method to do the work
     */
    private native void nativeOneDimensional(double in[], double out[], int dirn, int flag);


    /**   Method to take a Complex one-dimensional FFT with the real and imaging
     *    data split format in a two-dimensional array of size [2][length] 
     *    with the the i th component with real in the [0][i] element and the imagimary
     *    element in the [1][i] element.
     *    @param data two-dimensional array, real in [0][i] and imag in [0][i]
     *    @param dirn forward or inverse transform
     *    @param overwrite if true overwrite given data, else return in new space
     *    @return <code>double[][]</code> transformed data
     *    @throws <code>IllegalArgumentException</code> if real and imaginary array of different length.
     */
    public double[][] oneDimensional(double[][] data, int dirn, boolean overwrite) {
	//           Basic  sanity check
	if (data == null || data.length != 2 || data[0].length == 0 || data[1].length == 0) {
	    throw new NullPointerException("FFTComplex.oneDimensional: called with null or zero length data array(s).");
	}
	if ( data[0].length != data[1].length) {
	    throw new IllegalArgumentException("FFTWComplex.oneDimensional: called with real array of length: " 
					       + data[0].length + " and imaginary array of length: " + data[1].length);
	}
     
	double out[][] = null;            // The output array
	if (overwrite) {                  // Overwrite
	    out = data;
	}
	else {                            // Make new space
	    out = new double[2][data[0].length];
	}

	if (loaded) {
	    //          If forward send data in normal order
	    if (dirn == FORWARD) {
		nativeOneDimensionalSplit(data[0],data[1],out[0],out[1],getPlanFlag());
	    }
	    else {   // For inverse swap real/imag parts
		nativeOneDimensionalSplit(data[1],data[0],out[1],out[0],getPlanFlag());
	    }
	}
	else {
	    errorStream.println("FFTWComplex.oneDimensional: library not loaded.");
	}
	    
	return out;
    }

    /**     native OneDimensionalSplit to do the work
     */
    private native void nativeOneDimensionalSplit(double realin[], double imagin[],
						  double realout[], double imagout[],
						  int flag);
	

    /**      Method to take two-dimensional Complex FFT with the data supplied
     *       in a one-dimensional double array with real parts in even elements,
     *       and imaginary in the odd.
     *       Element i,j is located at 
     *       
     *       Real part     2*(j*width + i)
     *       Imag part     2*(j*width + i) + 1
     *
     *       @param width the width of the image data
     *       @param height the heigh of the image data
     *       @param data the data to be transformed
     *       @param dirn direction +1 for forward, -1 to backward
     *       @param overwrite if true data is overwritten with FFT, else a
     *              new array is retunded.
     *       @return <code>double[]</code> the FFTed array (if overwrite = true, this will
     *               be the same at the input data array.
     *       @throws <coce>IllegalArgumentException</code> if width and height does not match the array length.
     *
     */
    
    public double[] twoDimensional(int width, int height, double data[], int dirn, boolean overwrite) {
	//             basic sanity check
	if (data == null || data.length == 0) {
	    throw new NullPointerException("FFTComplex.twoDimensional: called with null or zero length data array.");
	}

	if (data.length != 2*width*height) {
            throw new IllegalArgumentException("FFTWComplex.twoDimensional: called with width: " + width +
                                               " height: " + height + " but data length of: " + data.length);
	}


	double out[] = null;                                 // The output array
	if (overwrite) {                                     // In place transform
	    out = data;
	}
	else {      
	    out = new double[data.length];                   // Make new output array
	}
	if (loaded) {
	    nativeTwoDimensional(width,height,data,out,dirn,getPlanFlag());
	}
	else {
	    errorStream.println("FFTWComplex.twoDimensional: library not loaded.");
	}
	    
	return out;
    }
    
    /**       nativeTwoDimensional method to do the work
     */
    private native void nativeTwoDimensional(int width, int height, double in[], double out[], int dirn, int flag);



    /**   Method to take a Complex two-dimensional FFT with the real and imaging
     *    data split format in a two-dimensional array of size [2][length] 
     *    with the the i th component with real in the [0][i] element and the imagimary
     *    element in the [1][i] element.
     *    @param width width of the data
     *    @param height height of the data
     *    @param data two-dimensional array, real in [0][i] and imag in [0][i]
     *    @param dirn forward or inverse transform
     *    @param overwrite if true overwrite given data, else return in new space
     *    @return <code>double[][]</code> transformed data
     *    @throws <code>IllegalArgumentException</code> if real and imaginary array of different length, or
     *                 width*height does not match the array lengths.
     */
    public double[][] twoDimensional(int width, int height, double[][] data, int dirn, boolean overwrite) {
	//           Basic  sanity check
	if (data == null || data.length != 2 || data[0].length == 0 || data[1].length == 0) {
	    throw new NullPointerException("FFTComplex.twoDimensional: called with null or zero length data array(s).");
	}
	if ( data[0].length != data[1].length || data[0].length != width*height) {
	    throw new IllegalArgumentException("FFTWComplex.twoDimensional: called with width : " + width +
					       " height: " + height + " but array length(s) of " +
					       data[0].length + " and " + data[1].length);
	}
     
	double out[][] = null;            // The output array
	if (overwrite) {                  // Overwrite
	    out = data;
	}
	else {                            // Make new space
	    out = new double[2][data[0].length];
	}

	if (loaded) {

	    //          If forward send data in normal order
	    if (dirn == FORWARD) {
		nativeTwoDimensionalSplit(width,height,data[0],data[1],out[0],out[1],getPlanFlag());
	    }
	    else {   // For inverse swap real/imag parts
		nativeTwoDimensionalSplit(width,height,data[1],data[0],out[1],out[0],getPlanFlag());
	    }
	}
	else {
	    errorStream.println("FFTWComplex.twoDimensional: library not loaded.");
	}
	
	return out;
    }
    
    /**     native TwoDimensionalSplit to do the work
     */
    private native void nativeTwoDimensionalSplit(int width, int height,
						  double realin[], double imagin[],
						  double realout[], double imagout[],
						  int flag);


    /**      Method to take three-dimensional Complex FFT with the data supplied
     *       in a one-dimensional double array with real parts in even elements,
     *       and imaginary in the odd.
     *
     *       Element (i,j,k) is located at:
     *
     *       RealPart  2*(k*width*height + j*width + i)
     *       ImagPart  2*(k*width*height * j*width + i) + 1
     *
     *       @param width the width of the image data
     *       @param height the heigh of the image data
     *       @param data the data to be transformed
     *       @param dirn direction +1 for forward, -1 to backward
     *       @param overwrite if true data is overwritten with FFT, else a
     *              new array is retunded.
     *       @return double[] the FFTed array (if overwrite = true, this will
     *               be the same at the input data array.
     *       @throws <coce>IllegalArgumentException</code> if width, height and depth does not match the array length.
     */
    
    public double[] threeDimensional(int width, int height, int depth, double data[], int dirn, boolean overwrite) {

	//             basic sanity check
	if (data == null || data.length == 0) {
	    throw new NullPointerException("FFTComplex.threeDimensional: called with null or zero length data array.");
	}

	if (data.length != 2*width*height*depth) {
            throw new IllegalArgumentException("FFTWComplex.threeDimensional: called with width: " + width +
                                               " height: " + height + " depth: " + depth +
					       " but data length of: " + data.length);
	}
	
	double out[] = null;                                 // The output array

	if (overwrite) {                                     // In place transform
	    out = data;
	}
	else {      
	    out = new double[data.length];                   // Make new output array
	}

	if (loaded) {
	    nativeThreeDimensional(width,height,depth,data,out,dirn,getPlanFlag());
	}
	else {
	   errorStream.println("FFTWComplex.threeDimensional: library not loaded.");
	} 
	return out;
    }
    
    /**       nativeTwoDimensional method to do the work
     */
    private native void nativeThreeDimensional(int width, int height, int depth, double in[], 
					       double out[], int dirn, int flag);



    /**   Method to take a Complex the-dimensional FFT with the real and imaging
     *    data split format in a two-dimensional array of size [2][length] 
     *    with the the i th component with real in the [0][i] element and the imagimary
     *    element in the [1][i] element.
     *    @param width width of the data
     *    @param height height of the data
     *    @param depth depth of the data
     *    @param data two-dimensional array, real in [0][i] and imag in [0][i]
     *    @param dirn forward or inverse transform
     *    @param overwrite if true overwrite given data, else return in new space
     *    @return <code>double[][]</code> transformed data
     *    @throws <code>IllegalArgumentException</code> if real and imaginary array of different length, or
     *                 width*height does not match the array lengths.
     */
    public double[][] threeDimensional(int width, int height, int depth, double[][] data, int dirn, boolean overwrite) {
	//           Basic  sanity check
	if (data == null || data.length != 2 || data[0].length == 0 || data[1].length == 0) {
	    throw new NullPointerException("FFTComplex.threeDimensional: called with null or zero length data array(s).");
	}
	if ( data[0].length != data[1].length || data[0].length != width*height*depth) {
	    throw new IllegalArgumentException("FFTWComplex.threeDimensional: called with width : " + width +
					       " height: " + height + " depth: " + depth +
					       " but array length(s) of " +
					       data[0].length + " and " + data[1].length);
	}
     
	double out[][] = null;            // The output array
	if (overwrite) {                  // Overwrite
	    out = data;
	}
	else {                            // Make new space
	    out = new double[2][data[0].length];
	}

	if (loaded) {
	    //          If forward send data in normal order
	    if (dirn == FORWARD) {
		nativeThreeDimensionalSplit(width,height,depth,data[0],data[1],out[0],out[1],getPlanFlag());
	    }
	    else {   // For inverse swap real/imag parts
		nativeThreeDimensionalSplit(width,height,depth,data[1],data[0],out[1],out[0],getPlanFlag());
	    }
	}
	else {
	    errorStream.println("FFTWComplex.threeDimensional: library not loaded.");
	} 
	return out;
    }
    
    /**     native TwoDimensionalSplit to do the work
     */
    private native void nativeThreeDimensionalSplit(int width, int height, int depth,
						  double realin[], double imagin[],
						  double realout[], double imagout[],
						  int flag);
    


}
    
    
